#include "nPresenter.h"

#include "widgets/WaitingIndicator.h"
#include "widgets/SectionViewer.h"
#include "widgets/PreviewCurrentPageWidget.h"
#include "widgets/PreviewNextPageWidget.h"
#include "widgets/ClockWidget.h"


nPresenter * nPresenter::_nPresenterInstance = 0;

nPresenter & nPresenter::instance()
{
	return *_nPresenterInstance;
}

nPresenter::nPresenter(QWidget *parent)
    : QMainWindow(parent)
{

	if(_nPresenterInstance)
	{
		qDebug() << "nPresenter is singleton...";
		exit(0);
	}
	_nPresenterInstance  = this;
	
	
	_layoutNames.append("1");
	_layoutNames.append("1:N");
	_layoutNames.append("N:1");
	_layoutNamesCurrent = 0;
	
	
	this->resize(1280, 750);
	this->setWindowState(Qt::WindowMaximized);
	this->setWindowTitle("nPresenter::Notes");
	
	this->setupWidgets();

	
	_updateCount = 0;
	
	_freezeEnabled = false;
	_blackoutEnabled = true;

	emit this->freezeEnabled(_freezeEnabled);
	emit this->blackoutEnabled(_blackoutEnabled);

	this->show();
	_presentationViewport.show();
	
	this->activateWindow();

}


void nPresenter::setupWidgets()
{

	
//	/* add slide preview */
	
	_previewCurrentDock.setFeatures(0);
	_previewCurrentDock.setWindowTitle("Current Page");
	_previewCurrentDock.setWidget(new PreviewCurrentPageWidget(this));
	this->addDockWidget(Qt::RightDockWidgetArea, &_previewCurrentDock);
	
	_previewNextDock.setFeatures(QDockWidget::DockWidgetMovable);
	_previewNextDock.setWindowTitle("Next Page");
	_previewNextDock.setWidget(new PreviewNextPageWidget());
	this->addDockWidget(Qt::RightDockWidgetArea, &_previewNextDock);
	
	/* notes viewport */
	_notesViewport.setSizePolicy(QSizePolicy::MinimumExpanding,
															 QSizePolicy::MinimumExpanding);
	this->setCentralWidget(&_notesViewport);	
	

	/* presentation viewport */
	_presentationViewport.setMinimumSize(480, 320);
	_presentationViewport.setWindowTitle("nPresenter::Presentation");


	/* clock widget */
	ClockWidget * clockWidget = new ClockWidget();
	clockWidget->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
	_clockDock.setWidget(clockWidget);
	_clockDock.setFeatures(QDockWidget::DockWidgetMovable | QDockWidget::DockWidgetClosable);
	_clockDock.setWindowTitle("Time");
	this->addDockWidget(Qt::LeftDockWidgetArea, &_clockDock);
	

	
	/* section viewer */
	
	QDockWidget * sectionViewerDock = new QDockWidget(this);
	
	SectionViewer * sV = new SectionViewer();
	sV->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::MinimumExpanding);
	sV->setMinimumSize(200, 100);
	sectionViewerDock->setWidget(sV);
	sectionViewerDock->setFeatures(QDockWidget::DockWidgetMovable);
	sectionViewerDock->setWindowTitle("Sections");
	this->addDockWidget(Qt::TopDockWidgetArea, sectionViewerDock);
	connect(sectionViewerDock,
					SIGNAL(dockLocationChanged(Qt::DockWidgetArea)),
					sV,
					SLOT(changeDockArea(Qt::DockWidgetArea)));
	connect(sectionViewerDock,
					SIGNAL(topLevelChanged(bool)),
					sV,
					SLOT(changeTopLevel(bool)));
	connect(this, SIGNAL(documentChanged()),
					sV, SLOT(updateSections()));
	connect(this, SIGNAL(currentPageChanged(int)),
					sV, SLOT(setActivePage(int)));
	
	
	this->setCorner(Qt::TopRightCorner, Qt::RightDockWidgetArea);
	this->setCorner(Qt::BottomRightCorner, Qt::RightDockWidgetArea);
	
	this->setCorner(Qt::TopLeftCorner, Qt::TopDockWidgetArea);
	this->setCorner(Qt::BottomLeftCorner, Qt::BottomDockWidgetArea);
	
	QWidgetList viewportWidgets = qApp->allWidgets();
	while(!viewportWidgets.empty())
	{
		ViewportWidget * widget = qobject_cast<ViewportWidget*>(viewportWidgets.takeFirst());

		if(!widget)
			continue;

		connect(this,
						SIGNAL(blackoutEnabled(bool)),
						widget,
						SLOT(enableBlackout(bool))
						);

		connect(this,
						SIGNAL(freezeEnabled(bool)),
						widget,
						SLOT(enableFreeze(bool))
						);
		connect(&_document,
						SIGNAL(requestedViewportReady(Viewports,int,QSize)),
						widget,
						SLOT(viewportIsReady(Viewports,int,QSize)));
		connect(this,
						SIGNAL(currentPageChanged(int)),
						widget,
						SLOT(requestPage(int)));
		
		connect(widget, SIGNAL(viewportSizeChanged(ViewportWidget*)),
						this, SLOT(viewportSizesChanged()));
	}

	_waitingIndicator = new WaitingIndicator();
	_waitingIndicator->setParent(this);
	_waitingIndicator->resize(100, 100);
	
	
	connect(this, SIGNAL(waitingOnPage()),
					_waitingIndicator, SLOT(show()));
	connect(&_document, SIGNAL(finishedRendering()),
					_waitingIndicator, SLOT(hide()));
	
	
}

nPresenter::~nPresenter()
{

}

Document & nPresenter::document()
{
	return instance()._document;
}

int nPresenter::currentPage()
{
	return instance()._currentPage;
}

QRectF nPresenter::viewport(Document::Viewports viewport)
{
	if(viewport == Document::NotesViewport)
		return QRectF(0.5, 0.0, 0.5, 1.0);
	if(viewport == Document::PresentationOneViewport)
		return QRectF(0.0, 0.0, 0.5, 1.0);
	
	return QRectF();
}


void nPresenter::mouseReleaseEvent(QMouseEvent *event)
{
	event->accept();
}

void nPresenter::keyReleaseEvent(QKeyEvent *event)
{
	/* key Right - next page */
	if(event->key() == Qt::Key_Right)
	{
		this->toNextPage();
	}
	
	/* key Left - previous page */
	if(event->key() == Qt::Key_Left)
	{
		this->toPreviousPage();
	}

	/* key Home - go to first page */
	if(event->key() == Qt::Key_Home)
		this->toPage(0);
	
	/* key B - toggle Blackout */
	if(event->key() == Qt::Key_B)
	{
		/* toggle blackout */
		_blackoutEnabled = !_blackoutEnabled;
		emit this->blackoutEnabled(_blackoutEnabled);
	}

	if(event->key() == Qt::Key_C)
	{
		_clockDock.setVisible(!_clockDock.isVisible());
	}
	
	/* key F - enable freeze */
	/* enable page freeze. the slide will not be changed on the presentation view */
	if(event->key() == Qt::Key_F)
	{
		_freezeEnabled = !_freezeEnabled;
		emit this->freezeEnabled(_freezeEnabled);
	}

	/* key O - open a document*/
	if(event->key() == Qt::Key_O)
	{

		/* ask for a file until a document could be created */
		bool success = false;
		while(!success)
		{
			QString fileName = QFileDialog::getOpenFileName(this,
																											tr("Open Presentation"),
																											QDir::home().path(),
																											tr("PDF Files (*.pdf)"));
			if(fileName.isNull())
				break;
			success = this->loadDocument(fileName);
		}
		
		if(!success)
			return;
		
		/* Check whether current page is inside document page bounds.
			 If so, set to current page.
			 So you can reload your document during presentation */
			 
		int newCurrentPage = _currentPage;
		if(newCurrentPage < 0 || newCurrentPage >= (_document.numPages()-1))
			newCurrentPage = 0;
		
		this->toPage(newCurrentPage);
		this->requestAllPages();
		
		/* clear the help text */
		_notesViewport.update();
	}

	/* key L - cycle through viewpot layouts */
	if(event->key() == Qt::Key_L)
	{ 
		_layoutNamesCurrent++;
		if(_layoutNamesCurrent >= _layoutNames.size())
			_layoutNamesCurrent = 0;
		document().setViewportLayout(_layoutNames[_layoutNamesCurrent]);
		this->requestAllPages();
	}


	event->accept();
}

bool nPresenter::loadDocument(QString path)
{
	
	if(!_document.open(path))
		return false;
	
	emit this->documentChanged();
	
	return true;
}

void nPresenter::showEvent(QShowEvent * event)
{
	event->accept();
}

void nPresenter::hideEvent(QHideEvent *event)
{
	event->accept();
	qApp->quit();
}

bool nPresenter::event(QEvent *event)
{
	
	/* react on layout changes */
	
//	if(event->type() == QEvent::MouseButtonRelease)
//	{
//			
//	}
	
	if(event->type() == QEvent::Resize || event->type() == QEvent::Paint)
	{
		QWidget * widget = this->centralWidget();
		
		/* if we have a window, center on it */
		if(widget)
		{
			QRect wRect = widget->geometry();
			QSize oSize = _waitingIndicator->size();
			_waitingIndicator->move(wRect.x() + wRect.width()/2 - oSize.width()/2, 
															wRect.y() + wRect.height()/2 -oSize.height()/2);
		}
		
//		if(this->isVisible() && event->type() == QEvent::Resize)
//			this->updatePages();
//
	}
	
	return QMainWindow::event(event);
}

void nPresenter::toPage(int pageNum)
{

	if(!this->document().loaded())
		return;

	/* page mus be inside page count boundaries */
	if(pageNum < 0)
		pageNum = 0;

	if(pageNum >= this->document().numPages())
		pageNum = this->document().numPages()-1;

	_currentPage = pageNum;

	/* stop any active rendering so that the new page will show up fast */
	document().abortRendering();

	/* finally signal that the current page has changed */
	emit this->currentPageChanged(_currentPage);
	
	if(!document().isPageRendered(this->currentPage()))
	{
		//document().abortRendering();
		emit waitingOnPage();
	}

		
}

void nPresenter::toNextPage()
{
	if(_currentPage+1 < this->document().numPages())
		this->toPage(_currentPage + 1);
}

void nPresenter::toPreviousPage()
{
	if(_currentPage-1 >= 0)
		this->toPage(_currentPage - 1);
}

void nPresenter::viewportSizesChanged()
{
	
	/* no document loaded. no need to update */
	if(!document().loaded())
		return;

	_updateCount++;
	QTimer::singleShot(500, this, SLOT(sizeChangeTimeout()));
}

void nPresenter::sizeChangeTimeout()
{
	_updateCount--;
	
	/* still updates active. */
	if(_updateCount > 0)
		return;
	
	this->requestAllPages();
	
}

void nPresenter::requestAllPages()
{
	this->document().abortRendering();
	this->document().clearCache();
	
	QList<QWidget*> widgets = qApp->allWidgets();
	//QList<QWidget*> widgets = this->findChildren<QWidget*>();
	int pageCount = document().numPages();
	
	/* first request the current page */
	for(int i = 0; i < widgets.size(); i++)
	{	
		ViewportWidget * widget = qobject_cast<ViewportWidget*>(widgets[i]);
		if(!widget)
			continue;
		widget->requestPage(this->currentPage());
	}
	
	/* then do the rest */
	while(!widgets.empty())
	{
		ViewportWidget * widget = qobject_cast<ViewportWidget*>(widgets.takeFirst());
		
		if(!widget)
			continue;
		
		for(int i = 0; i < pageCount; i++)
			widget->requestPage(i);
	}

	emit waitingOnPage();
}
