#ifndef NPRESENTER_H
#define NPRESENTER_H

#include <QtCore/QtCore>
#include <QtGui/QtGui>

#include <QtGui/QMainWindow>

#include "pdf/Document.h"
#include "widgets/NotesViewportWidget.h"
#include "widgets/PresentationViewWidget.h"
#include "widgets/WaitingIndicator.h"


/** Main Window for Application
 * This class handles the main part of the Presenter.
 * It holds all Widgets and does user interaction by event() and keyReleaseEvent()
 * @nosubgrouping
 */
class nPresenter : public QMainWindow
{
    Q_OBJECT
		
		/* singleton instance pointer */
		static nPresenter * _nPresenterInstance;
		
		int _updateCount;
		
		/* widgets */
		QDockWidget _treeDock;
		QDockWidget _previewCurrentDock;
		QDockWidget _previewNextDock;
		QDockWidget _clockDock;
		QDockWidget _notesDock;
		NotesViewportWidget _notesViewport;
		PresentationViewportWidget _presentationViewport;
		WaitingIndicator * _waitingIndicator;
		
		/* presentation properties */
		Document _document;
		QFutureWatcher<void> _futureWatcher;
		QFuture<void> _renderFuture;
		bool _blackoutEnabled;
		bool _freezeEnabled;
		int _currentPage;
		
		QStringList _layoutNames;
		int _layoutNamesCurrent;

	public:
    nPresenter(QWidget *parent = 0);
    virtual ~nPresenter();
		
		/** get the singleton instance of nPresenter */
		static nPresenter & instance();
		
		/** PDF Abstraction Instance */
		static Document &  document();
		
		/** current active Page */
		static int currentPage();
		
		
		
		/** gets geometry for viewport
		 * @returns the viewport geometry or an invalid QRectF if viewport is not defined
		 */
		static QRectF viewport(Document::Viewports viewport);

	protected:	
		void mouseReleaseEvent(QMouseEvent * event);
		void keyReleaseEvent(QKeyEvent * event);
		void showEvent(QShowEvent * event);
		void hideEvent(QHideEvent * event);
		bool event(QEvent *event);
		
		
	private:
		/** setup widget and connect signals/slots */
		void setupWidgets();
		
	signals:
		
		/** the current page changed */
		void currentPageChanged(int numPage);
		
		/** work in progress */
		void waitingOnPage();
		
		/** new document loaded */
		void documentChanged();

		/** blackout status changed */
		void blackoutEnabled(bool enabled);
		/** page freeze status changed */
		void freezeEnabled(bool enabled);
		
	private slots:
		
		void sizeChangeTimeout();
		
	
	public slots:
		
		/** loads a PDF file */
		bool loadDocument(QString path);
		
		/** called from ViewportWidgets when resized
		 * Defer update of viewports for a while.
		 * This prevents constantly updating viewports while resizing ViewportWidget
		 */
		void viewportSizesChanged();
		
		
		/** request alls pages and viewports again
		 * clears Document's image cache and requests all pages.
		 * @deprecated : should be replaced by a more intelligent caching algorithm
		 */
		void requestAllPages();
		
		/** @name Page Navigation */
		/** @{ */
		/** handles page changes */
		void toPage(int pageNum);
		
		/** go to next page
		 * @see toPage()
		 */
		void toNextPage();
		
		/** go to previous page
		 * @see toPage()
		 */
		void toPreviousPage();
		/** @} */
};

#endif // NPRESENTER_H
