#include "pdf/Document.h"
#include "nPresenter.h"

Document::Document(QObject * parent) :
    QThread(parent), _mutex(QMutex::Recursive)
{
	
	qRegisterMetaType<Viewports>("Viewports");
	_doc = 0;
	
	/* move this QObject to own thread */
	this->moveToThread(this);
	
	
	/* initialize Viewport Layouts */
	
	/* presentation has one viewport - the presentation itself */
	_viewportLayouts["1"][NotesViewport] = QRectF(0.0, 0.0, 1.0, 1.0);
	_viewportLayouts["1"][PresentationOneViewport] = QRectF(0.0, 0.0, 1.0, 1.0);

	/* left half: presentation; right half - notes */
	_viewportLayouts["1:N"][NotesViewport] = QRectF(0.5, 0.0, 0.5, 1.0);
	_viewportLayouts["1:N"][PresentationOneViewport] = QRectF(0.0, 0.0, 0.5, 1.0);
	
	/* left half: notes; right half: presentation */
	_viewportLayouts["N:1"][NotesViewport] = QRectF(0.0, 0.0, 0.5, 1.0);
	_viewportLayouts["N:1"][PresentationOneViewport] = QRectF(0.5, 0.0, 0.5, 1.0);
	
	/* start event loop */
	this->start();
}

Document::~Document()
{
	/* stop thread */
	this->quit();
	this->wait();
}

bool Document::open(QString path)
{
	try
	{
		/* openign a a document should be done inside the rendering thread.
			 if we are called from within the own thread, call the opening function directly.
			 if the call comes from another thread, "BlockedQueued"-call the function.
			 This distinction is needed because a BlockingQueuedConnection deadlocks
			 when called from the same thread.
		*/
//		if(QThread::currentThread() == this->thread())
//		{	
			this->openPdfDocument(path);
//		}
//		else
//		{
//			QMetaObject::invokeMethod(this, "openPdfDocument", Qt::BlockingQueuedConnection,
//																Q_ARG(QString, path));
//		}
	}
	catch(...)
	{
		qDebug() << "Error while opening Document " << path;
		return false;
	}
	return true;
}

bool Document::loaded()
{
	if(_doc)
		return true;
	return false;
}

bool Document::isPageRendered(int page)
{
	QMutexLocker locker(&_mutex);
	return _pages.contains(page);
}

void Document::setPdfDocument(Poppler::Document *pdfDoc)
{
	
	if(_doc)
		delete _doc;
	
	_doc = pdfDoc;
	_doc->setRenderHint(Poppler::Document::TextAntialiasing, true);
	_pageCount = _doc->numPages();
	
	
	/* extract the Keywords Meta Info
		 This program uses this string to setup the ViewportLayout */
	QString viewportLayout = _doc->info("Keywords");
	
	if(viewportLayout.isNull())
		viewportLayout = "1";
	
	if(_viewportLayouts.contains(viewportLayout))
		_viewports = _viewportLayouts.value(viewportLayout);
	
	if(_viewports.empty())
	{
		_viewports = _viewportLayouts.value("1");
		qDebug() << "Document::setPdfDocument(): Could not extract meaningfull Viewportlayout.";
	}
	
}

void Document::openPdfDocument(QString path)
{
	QMutexLocker locker(&_mutex);
	Poppler::Document * pdfDoc = Poppler::Document::load(path);
	if(!pdfDoc)
	{
		qDebug() << "Loading PDF:" << path << ": failed.";
		throw QString("Error loading document");
	}
	
	this->abortRendering();
	this->clearCache();
	this->setPdfDocument(pdfDoc);
}

void Document::setViewportLayout(QString name)
{
	QMutexLocker locker(&_mutex);
	if(!_viewportLayouts.contains(name))
		return;
	_viewports = _viewportLayouts.value(name);
	this->abortRendering();
	this->clearCache();
}

QRectF Document::viewportGeometry(Document::Viewports viewport)
{
	if(!_viewports.contains(viewport))
		return QRectF();
	
	return _viewports.value(viewport);
}

int Document::numPages()
{
	
	return _pageCount;
}

QDomDocument * Document::toc()
{
	
	QMutexLocker locker(&_mutex);
	QDomDocument * doc = _doc->toc();
	
	if(doc)
		return doc;

	/* There is no table of content.
		 Create a pseudo TOC by enumerating the pages */
	doc = new QDomDocument("sections");
	
	int pageNum = this->numPages();
	
	for(int i = 0; i < pageNum; i++)
	{
		QString sectionName = "Page " + QString::number(i+1);
		QDomElement tmpElem = doc->createElement(sectionName);
		doc->appendChild(tmpElem);
		
		/* Poppler is quite undocumented here.
			 Destination is a semicolon-seperated enumeration of properties.
			 Second argument is the Pagenumber.
		*/
		tmpElem.setAttribute("Destination", "2;"+QString::number(i+1));
	}
	
	return doc;
}




/** @todo cache results into a hash to make it really threadsafe */
int Document::pageNumberByBookmark(QString bookmark)
{
	
	if(bookmark.contains("DestinationName/"))
	{
		QMutexLocker locker(&_mutex);
		bookmark = bookmark.replace("DestinationName/", "", Qt::CaseSensitive);
		Poppler::LinkDestination * dest = _doc->linkDestination(bookmark);
	
		if(!dest)
			return -1;
		int pageNum = dest->pageNumber();
		delete dest;
		
		return pageNum;
	}
	
	bookmark = bookmark.replace("Destination/", "", Qt::CaseSensitive);
	QList<QString> seps = bookmark.split(";");
	if(seps.length() >= 2)
		return seps[1].toInt();
	return 0;
	
}

QImage Document::viewport(Viewports viewport, int page, QSize targetSize)
{
	QMutexLocker locker(&_mutex);
	
	if(!this->isViewportAvailable(viewport, page, targetSize))
		return QImage();
		
	return _pages[page][viewport].image(targetSize);
}

bool Document::isViewportAvailable(Viewports viewport, int page, QSize targetSize)
{
	QMutexLocker locker(&_mutex);
	
	if(!_pages.contains(page))
		return false;
	if(!_pages[page].contains(viewport))
		return false;
	if(!_pages[page][viewport].contains(targetSize))
		return false;
	
	return true;
}

bool Document::requestViewport(Viewports viewport, int page, QSize targetSize)
{
	QMutexLocker locker(&_mutex);
	
	/* Check if we support the given viewport.
	   This can be if not all Document::Viewports are configured. */
	if(this->viewportGeometry(viewport).isNull())
		return false;
	
	/* Hurray! The page is already rendered. Dont start to render but
		 tell that the requested viewport is already ready. */
	if(this->isViewportAvailable(viewport, page, targetSize))
	{
		emit requestedViewportReady(viewport, page, targetSize);
		return true;
	}
	
	/* Append a new RenderRequest to the queue. */
	_renderRequests.append(RenderRequest(page, targetSize, viewport));
	
	/* Currently nothing is rendered, so initiate it.
     This function is thread-safe. Therefor invokeMethod()
     is needed to run Document::renderRequest().
     inside Document's thread. */
	if(_renderRequests.size() < 2)
		QMetaObject::invokeMethod(this, "renderRequest", Qt::QueuedConnection);
	
	return true;
			
}

void Document::abortRendering()
{
	QMutexLocker locker(&_mutex);
	_renderRequests.clear();
}

void Document::clearCache()
{
	QMutexLocker locker(&_mutex);
	_pages.clear();
}

void Document::renderRequest()
{
	Q_ASSERT(QThread::currentThread() == this->thread());
	
	QMutexLocker locker(&_mutex);
	
	/* nothing to do */
	if(_renderRequests.empty())
		return;
	
	RenderRequest request = _renderRequests.takeFirst();
	
	/* From now on we don't need to protect anything. 
	   Allow the GUI-thread to place new rendering requests or
		 or abort them. */
	locker.unlock();
	
	QRectF viewport = this->viewportGeometry(request._viewport);
	int page = request._page;
	QSize targetSize = request._targetSize;
	
	QImage rendered;
	
	Poppler::Page * pdfPage = 0;
	
	try {
		
		
		if(!_doc)
			throw QString("No Document loaded.");
		
		if(page < 0 || page >= _pageCount)
			throw QString("Page number is out of bounds 0..."+QString::number(_pageCount));
		
		pdfPage = _doc->page(page);
		
		if(!pdfPage)
			throw QString("Page not found.");
		
		QSizeF pageSize = pdfPage->pageSizeF();
		
		/* Calculate the resolution we need to fit the rendered
			 viewport into targetSize.
			 pageSize is in 1/72 inch.
			 xres and yres should be in dots per inch */
		qreal xres = targetSize.width()/( pageSize.width()*viewport.width() ) *72.0;
		qreal yres = targetSize.height()/( pageSize.height()*viewport.height() ) * 72.0;
	
		
		/* calculate the beginning of the viewport
			 in documents coordinate (1/72 inch) */
		int vLeft = qRound((viewport.x()*pageSize.width()/72.0)*xres);
		int vTop = qRound((viewport.y()*pageSize.height()/72.0)*yres);
		
		/* calculate the width and height of the viewport 
			 in documents coordinates ( 1/72 inch) */
		int vWidth = qRound((viewport.width()*pageSize.width()/72.0)*xres);
		int vHeight = qRound((viewport.height()*pageSize.height()/72.0)*yres);
		
		/* finally render the image to our buffer */
		rendered = pdfPage->renderToImage(xres, yres, vLeft, vTop, vWidth, vHeight);
		
		if(rendered.size() != targetSize)
		{
			qDebug() << "Document::renderRequest(): Requested size does not fit to rendered image size.";
			rendered.scaledToWidth(targetSize.width());
		}
		
		
	}
	catch(QString & err)
	{
		qDebug() << "Document::renderRequest(): Error while rendering Page: " << err;
	}
	catch(...)
	{
		qDebug() << "Document::renderRequest(): Error while rendering Page.";
	}
	
	/* Pages returned by Poppler::Document are owned by caller,
		 so delete it. */
	if(pdfPage)
		delete pdfPage;
	
	locker.relock();
	
	/* put result into render cache */
	if(!rendered.isNull())
		_pages[request._page][request._viewport].insert(rendered);
	
	/* Announce that we finished a viewport rendering */
	emit this->requestedViewportReady(request._viewport, request._page, request._targetSize);
	
	/* Check whether there are any more RenderRequests. If not, stop processing. */
	if(_renderRequests.isEmpty())
	{
		emit finishedRendering();
		return;
	}
	
	/* There are more request so process them. */
	QMetaObject::invokeMethod(this, "renderRequest", Qt::QueuedConnection);
}
