#ifndef DOCUMENT_H
#define DOCUMENT_H

#include <QtCore/QtCore>
#include <QtGui/QtGui>
#include <poppler/qt4/poppler-qt4.h>

/** PDF rendering and page caching 
 * This class abstract a PDF Document (Poppler::Document)
 * and does the page rendering in a seperate thread.
 * @note Using libpoppler with threads is pain-in-the-ass.
 */
class Document : public QThread
{
		Q_OBJECT
		Q_ENUMS(Viewports)
		
	public:
		
		/** defines possible viewports */
		enum Viewports {NotesViewport,
										PresentationOneViewport, 
										PresentationTwoViewport,
										PresentationThreeViewport};		
				
		
		/** render request */
		struct RenderRequest
		{
			int _page;
			QSize _targetSize;
			Document::Viewports _viewport;
			
			
			RenderRequest(int page, QSize targetSize, Document::Viewports viewport)
			{
				_page = page;
				_targetSize = targetSize;
				_viewport = viewport;
			}
			
			bool operator==(RenderRequest & other)
			{
				if(other._page == _page &&
					 other._targetSize == _targetSize &&
					 other._viewport == _viewport)
					return true;
				return false;
			}
		};
		
		/** Image cache helper class */
		class ImageMap
		{
			
			QList<QImage> _images;
			
			private:
				
				/** @brief find the index of the image for given size
				 * @returns the index or -1 if not found 
				 */
				int indexForSize(QSize size)
				{
					for(int i = 0; i < _images.size(); i++)
						if(_images[i].size() == size)
							return i;
					return -1;
				}
				
			public:
				
				/** 
				 *	@returns the image with given size or a null-Image if none found.
				 */			
				QImage image(QSize size)
				{
					int index = this->indexForSize(size);
					
					if(index < 0)
						return QImage();
					return _images[index];
				}
	
				/** @brief removes image with size from the map */
				void remove(QSize size)
				{
					int index = this->indexForSize(size);
					
					if(index < 0)
						return;
				}
				
				/** @brief checks whether image with size is available */
				bool contains(QSize size)
				{
					if(this->indexForSize(size) < 0)
						return false;
					return true;
				}	
				
				/** @brief inserts given image
				 * @note if image with same size is available, the image will be overwritten
				 */
				void insert(QImage image)
				{
					/* dont insert if the image is invalid */
					if(image.isNull())
						return;
					
					int index = this->indexForSize(image.size());
					
					/* image with image.size() available, delete first */
					if(index > 0)
						_images.removeAt(index);
						
					_images.append(image);
				}
			
				/** @brief clears ImageMap */
				void clear()
				{
					_images.clear();
				}
				
				bool isNull()
				{
					return _images.isEmpty();
				}
			
		};
		
		Poppler::Document * _doc;
		QMutex _mutex;
		int _pageCount;
		
		/** represents QHash<Page Number, Image Cache> */
		QHash<int, QMap<Viewports, ImageMap> > _pages;
		
		QList<RenderRequest> _renderRequests;
		
		/* store viewport configuration */
		QHash<Viewports, QRectF> _viewports;
		
		QHash<QString, QHash<Viewports, QRectF> > _viewportLayouts;
		
	public:
		
		explicit Document(QObject * parent = 0);
		
		virtual ~Document();
		
		/** @brief opens a PDF document by given path
		 * @threadsafe
		 */
		bool open(QString path);

		/** @brief number of pages in document
		 * @threadsafe
		 */
		int numPages();
		
		/** sets viewport layout to name */
		void setViewportLayout(QString name);
		
		/** @brief generates a table of content */
		QDomDocument * toc();
		
		/** @brief get page by bookmark name 
		 * Destinations given in toc() refer to certain viewports.
		 * This function returns the page on which the destination points on
		 * @threadsafe
		 */
		int pageNumberByBookmark(QString bookmark);
		
		/** @returns the viewport of page rendere in targetSize 
		 * @note if there is no such viewport rendered, a null image will be returned
		 * @threadsafe
		 */
		QImage viewport(Viewports viewport, int page, QSize targetSize);
		
		
		/** @brief checks if a document is loaded
		 * @threadsafe
		 */
		bool loaded();
		
		/** @brief checks if there are any entries for page
		 * @threadsafe
		 */
		bool isPageRendered(int page);
		
		/** gets geometry for viewport
		 * @returns the viewport geometry or an invalid QRectF if viewport is not defined
		 * @threadsafe
		 */
		QRectF viewportGeometry(Document::Viewports viewport);
		
	private:
		void setPdfDocument(Poppler::Document * pdfDoc);
		Q_INVOKABLE void openPdfDocument(QString path);
		
		/** @brief checks whether a given viewport is available
		 * @threadsafe
		 */
		bool isViewportAvailable(Viewports viewport, int page, QSize targetSize);
		
	signals:
		/** @brief emitted when a Viewport requested by requestViewport() is ready */
		void requestedViewportReady(Viewports viewport, int page, QSize targetSize);
	
		void finishedRendering();
		
	public slots:
		
		/** @brief request a given Viewport to be rendered
		 * @threadsafe
		 */
		bool requestViewport(Viewports viewport, int page, QSize targetSize);
		
		/** @brief stops the rendering
		 * @threadsafe
		 */
		void abortRendering();
		
		/** clears cache
		 * @threadsafe
		 */
		void clearCache();
		
	private slots:
		/** @brief renders page with given viewport into targetSize'd image */
		void renderRequest();
		


};
Q_DECLARE_METATYPE(Document::Viewports)

#endif // DOCUMENT_H
