#include "SectionViewer.h"
#include "nPresenter.h"
#include "pdf/Document.h"
#include "widgets/SectionViewer/SectionWidget.h"

SectionViewer::SectionViewer(QWidget *parent) :
    QWidget(parent)
{
	this->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Expanding);
	_orientation = Qt::Horizontal;
	_currentPage = -1;
	_toc = 0;
}

Qt::Orientation SectionViewer::orientation()
{
	return _orientation;
}

void SectionViewer::setActivePage(int pageNum)
{
	_currentPage = pageNum;
	for(int i = 0;  i < _sections.size(); i ++)
		_sections[i]->setActivePage(pageNum);
	this->layoutSections();
}

void SectionViewer::paintEvent(QPaintEvent *event)
{


	int gradHeight = 5;
	
	QPainter painter;
	painter.begin(this);
	painter.setRenderHint(QPainter::Antialiasing);
	painter.setClipRect(event->rect());
	//painter.fillRect(0,0,this->width(), this->height()-gradHeight, Qt::black);


	QLinearGradient grad;
	grad.setColorAt(0.0, QColor(0,0,0,140));
	grad.setColorAt(1.0, QColor(0,0,0,220));

	/* align gradient according to widget orientation */
	if(this->orientation() == Qt::Horizontal)
	{
		grad.setStart(QPointF(0.0, 0.0));
		grad.setFinalStop(QPoint(0.0, this->height()));
	}
	else
	{
		grad.setStart(QPointF(0.0, 0.0));
		grad.setFinalStop(QPoint(this->width(),0.0));
	}

	QBrush flowBrush(grad);


	QPainterPath path;

	path.addRoundedRect(gradHeight, /* center horizontal */
											gradHeight, /* center vertical */
											this->width()-2*gradHeight,
											this->height()-2*gradHeight, 
											5, 5); 
	painter.fillPath(path, flowBrush);

	painter.end();
}

QSize SectionViewer::sizeHint() const
{
	return QSize(0, 0);
}

void SectionViewer::resizeEvent(QResizeEvent *event)
{
	event->accept();
	this->layoutSections();
}


void SectionViewer::changeTopLevel(bool topLevel)
{
	topLevel = topLevel;
	//	if(topLevel || this->orientation() == Qt::Vertical)
//		_orientation = Qt::Horizontal;
//	this->layoutSections();
}

void SectionViewer::changeDockArea(Qt::DockWidgetArea area)
{
	_orientation = Qt::Horizontal;
	if(area == Qt::TopDockWidgetArea || area == Qt::BottomDockWidgetArea)
		_orientation = Qt::Horizontal;
	if(area == Qt::LeftDockWidgetArea || area == Qt::RightDockWidgetArea)
		_orientation = Qt::Vertical;
	this->layoutSections();
}

void SectionViewer::layoutSections()
{
	
	int width = this->width();
	int height = this->height();

	/* resolution in dots per centimeter */
	qreal dpiY = qreal(this->physicalDpiY())/2.54;
	
	/* set divisionHeight to  a half centimeter on the display device */
	int divisionHeight = dpiY*0.5;

	this->setMinimumHeight(divisionHeight*4+15);
	this->setMinimumWidth(200);
	
	int divisionWidth = 0;
	
	QSize	sectionSize;

	int horizontalMargin = 5; /* px */

	if(this->orientation() == Qt::Horizontal)
	{
		divisionWidth = (width-3*2*horizontalMargin)/3;
		sectionSize.setHeight(this->height()-10);
		sectionSize.setWidth(divisionWidth);
	}
	else
	{
		divisionWidth = width-2*horizontalMargin;
		sectionSize.setWidth(width);
		sectionSize.setHeight(height/3);
	}

	QList<SectionWidget*> sections = this->findChildren<SectionWidget*>();
	while(!sections.isEmpty())
	{
		SectionWidget * widget = sections.takeFirst();
		widget->changeDivisionSize(QSize(divisionWidth,divisionHeight));
		widget->resize(sectionSize);
	}
	
	/* if we have no sections we have nothing to do */
	if(_sections.empty())
		return;
	
	sections = this->findChildren<SectionWidget*>();
	/* search for active section */
	int activeSection = 0;
	for(int i = 0; i < _sections.size(); i++)
	{
		if(_sections[i]->containsPage(_currentPage))
			activeSection = i;
		
		_sections[i]->setVisible(false);
	}
	


	/* no page set. so dont render anything */
	if(_currentPage < 0)
		return;
	
	int sectionWidth = this->width()/3;
	int sectionHeight = this->height()/3;
	
	if(this->orientation() == Qt::Horizontal)	
	{
		if(activeSection > 0)
		{
			_sections[activeSection-1]->move(5, 5);
			_sections[activeSection-1]->setVisible(true);
		}
	
			_sections[activeSection]->move(5 + sectionWidth, 5);
			_sections[activeSection]->setVisible(true);
	
		if((activeSection+1) < (_sections.size()))
		{
			_sections[activeSection+1]->move(5 + sectionWidth*2, 5);
			_sections[activeSection+1]->setVisible(true);
		}
	}
	else
	{
		if(activeSection > 0)
		{
			_sections[activeSection-1]->move(5, 5);
			_sections[activeSection-1]->setVisible(true);
		}
	
			_sections[activeSection]->move(5, 5 + sectionHeight);
			_sections[activeSection]->setVisible(true);
	
		if((activeSection+1) < (_sections.size()))
		{
			_sections[activeSection+1]->move(5, 5 + sectionHeight*2);
			_sections[activeSection+1]->setVisible(true);
		}	
	}
		
	this->update();
}

void SectionViewer::updateSections()
{

	while(!_sections.empty())
	{
		SectionWidget * widget = _sections.takeFirst();
		delete widget;
	}


	if(_toc)
		delete _toc;
	

	_toc = nPresenter::document().toc();

	if(!_toc)
		return;

	QDomElement tmpElem = _toc->firstChildElement();
	while(!tmpElem.isNull())
	{

		SectionWidget * widget = new SectionWidget(tmpElem, this);
		if(!widget)
			continue;
		_sections.append(widget);
		tmpElem = tmpElem.nextSiblingElement();
	}
	
	this->layoutSections();
	
}
